/*

The maximum clock rate is 20 MHz. The minimum clock rate is a function of the number of bits of data per
frame that need to be retrieved. As described in the sections that follow, the number of bits of data varies
depending upon user settings (video format mode, telemetry mode). 

For default conditions (Raw14 mode,
telemetry disabled), there are 60 video packets per frame, each 1312 bits long, at approximately 25.9 frames
per second. Therefore, the minimum rate is on the order of 2 MHz.


*/


module LAPTON(
output reg [6:0] line_index,
output reg [6:0] line_count,
output reg [15:0] data,
output reg [15:0] ID,
output reg [15:0] CRC,
output reg [15:0] frame_count,

	input 	clk,
	input		reset_n,
	
	// memory 
	output [15:0] mem_data,
	output [12:0]  mem_addr,
	output        mem_write,
	output        mem_clk,
	
	// 
	inout 	I2C_SCL,
	inout		I2C_SDA,
	output   SPI_CS_n,
	output	SPI_MOSI,
	input		SPI_MISO,
	output	SPI_CLK  // 2~20 MHz
);


assign SPI_MOSI = 1'b0;

`define BIT_NUM_PER_LINE	1312
`define LINE_NUM_PER_FRAME	60
`define IDLE_DUR	30


//////////////////////////////////
// write data to memory
assign mem_clk = clk;
assign mem_data = video_data;
assign mem_write = video_data_write;
assign mem_addr = video_address;


reg [15:0]	video_data;
reg 			video_data_write;
reg [12:0]	video_address;

always @(posedge mem_clk or negedge reset_n)
begin
	if (~reset_n)
		video_data_write <= 1'b0;
	else // if (xcvr_frame)
	begin
		if (is_word_ready && (bit_index+1 >= 3*16) && valid_line_data)
		begin
			video_data_write <= 1'b1;
			video_data <= data_new;
//			video_data <= line_index * 80 + (bit_index-2*16)/16;
			video_address <= line_index * 80 + (bit_index-2*16)/16;
		end
		else
			video_data_write <= 1'b0;
	end
	//else
	//	video_data_write <= 1'b0;
end





////////////////////////////////
// state machine

// major state
`define ST_FRAME_WAIT	0
`define ST_FRAME_START	1
`define ST_FRAME_XCVR	2
`define ST_FRAME_END		3

// line state
`define LINE_WAIT			0	
`define LINE_START		1	
`define LINE_READ_ID		2
`define LINE_READ_CRC	3
`define LINE_READ_DATA	4
`define LINE_DONE			5


// frame strigger
`define FRAME_INTERVAL	15000000/27   // 1/27 sec

reg [23:0] cnt;
always @(posedge clk or negedge reset_n)
begin
	if (~reset_n)
		cnt <= 0;
	else if (cnt < `FRAME_INTERVAL)
		cnt <= cnt + 1;
	else if (state == `ST_FRAME_START)
		cnt <= 0;
end

wire is_can_get_frame;
assign is_can_get_frame = (cnt >= `FRAME_INTERVAL)?1'b1:1'b0;


// major state
reg [1:0] state;
always @(posedge clk or negedge reset_n)
begin
	if (~reset_n)
		state <= `ST_FRAME_WAIT;
	else 
	begin
		case (state)
			`ST_FRAME_WAIT:
					if (is_can_get_frame)
					state <= `ST_FRAME_START;	
				
			`ST_FRAME_START:
				state <= `ST_FRAME_XCVR;
				
			`ST_FRAME_XCVR:
					if (line_state == `LINE_DONE && (line_count == `LINE_NUM_PER_FRAME)) // 
						state <= `ST_FRAME_END;				
				
			`ST_FRAME_END:
				state <= `ST_FRAME_WAIT;
				
		endcase
	end
end

// frame count
always @(posedge clk or negedge reset_n)
begin
	if (~reset_n)
		frame_count <= 0;
	else if (state == `ST_FRAME_END)
		frame_count <= frame_count + 1;
end


// line state
reg [2:0] line_state;
reg [5:0] line_wait_count;
always @(posedge clk)
begin
	if (state == `ST_FRAME_XCVR)
	begin
		case(line_state)
			`LINE_WAIT:
					if (line_wait_count < `IDLE_DUR)
						line_wait_count <= line_wait_count + 1;
					else 
						line_state <= `LINE_START;
				
			`LINE_START:
				line_state <= `LINE_READ_ID;
				
			`LINE_READ_ID:	
				if (is_word_ready)
					line_state <= `LINE_READ_CRC;
				
			`LINE_READ_CRC: 	
				if (is_word_ready)
					line_state <= `LINE_READ_DATA;	
				//begin
					//if ((ID[13:8] == 4'hF)  || (ID[7:0] >= `LINE_NUM_PER_FRAME) || ((ID[7:0] != 0) && ((PRE_ID[7:0]+1) != ID[7:0]))) //>= `LINE_NUM_PER_FRAME) || ((PRE_ID[7:0]+1) != ID[7:0]))
						//line_state <= `LINE_BAD;			
					//else
						//line_state <= `LINE_READ_DATA;					
				//end
				
			`LINE_READ_DATA:	
					if ((bit_index+1) == `BIT_NUM_PER_LINE)
						line_state <= `LINE_DONE; 				
			`LINE_DONE:	
					begin
						line_state <= `LINE_WAIT; 	
						line_wait_count <= 0;			
					end
		endcase
	end
	else
	begin
		line_state <= `LINE_WAIT;
		line_wait_count <= 0;
	end
end


///////////////////////////////////////
// read line

assign SPI_CS_n = ((state == `ST_FRAME_XCVR) && (line_state >= `LINE_START))?1'b0:1'b1;
assign SPI_CLK = (~SPI_CS_n && (line_state <= `LINE_READ_DATA))?clk:1'b1;

// bit index
reg [11:0] bit_index;
always @(posedge clk)
begin
	if (state == `ST_FRAME_XCVR)
	begin
		if (line_state == `LINE_WAIT)
			bit_index <= 0;
		else if ((line_state < `LINE_DONE) && (bit_index < `BIT_NUM_PER_LINE))
			bit_index <= bit_index + 1;
	end
	else
		bit_index <= 0;
end

// line index
//reg [11:0] lindex_index;
//always @(posedge clk)
//begin
//	if (state == `ST_FRAME_XCVR)
//	begin
//		if ((line_state == `LINE_READ_ID) && is_word_ready)
//			line_index <= data_new[11:0];
//	end
//	else
//		line_index <= 0;
//end

wire is_valid_id;

assign is_valid_id = ((ID[7:0] < `LINE_NUM_PER_FRAME) && (ID[7:0] == line_count)) ?1'b1:1'b0;


reg valid_line_data;
always @(posedge clk)
begin
	if (state == `ST_FRAME_XCVR)
	begin
		if ((line_state == `LINE_READ_CRC) && is_word_ready)
		begin
			valid_line_data <= is_valid_id;
			
			if (is_valid_id)
			begin
				line_count <= line_count + 1;
				line_index <= ID[7:0];
			end
			else
				line_count <= 0;
		end
	end
	else
	begin
		valid_line_data <= 1'b0;
		line_index <= 0;
		line_count <= 0;
	end
end

// read word
wire [15:0] data_new;
reg [15:0] data_queue;

assign data_new = {data_queue[14:0], SPI_MISO};
always @(posedge clk)
begin
	if (state == `ST_FRAME_XCVR)
	begin
		data_queue <= data_new;
	end
end

// parsing word
wire is_word_ready;
assign is_word_ready = ((state == `ST_FRAME_XCVR) && ((bit_index & 12'h00F) == 12'h00F))?1'b1:1'b0;
reg [15:0]	PRE_ID;

always @(posedge clk)
begin
	if (is_word_ready)
	begin
		if (line_state == `LINE_READ_ID)
		begin
			PRE_ID <= ID;
			ID <= data_new;
		end
		else if (line_state == `LINE_READ_CRC)
		begin
			CRC <= data_new;
		end
		else if (line_state == `LINE_READ_DATA)
		begin
			data <= data_new;
		end
	end
end


/*

//////////////////////////////////
// SPI_CLK & SPI_CS control

reg [5:0] idle_count;
wire spi_clk_en;
assign spi_clk_en = (idle_count < `IDLE_DUR)?1'b0:1'b1;


assign SPI_MOSI = 1'b0; // The MOSI (Master Out/Slave In) signal is not currently employed and should be grounded. 
assign SPI_CLK = spi_clk_en?clk:1'b1;
assign SPI_CS_n = (spi_clk_en | (pre_spi_clk_en & ~spi_clk_en))?1'b0:1'b1;

always @(posedge clk or negedge reset_n)
begin
	if (~reset_n)
		idle_count <= 0;
	else if (idle_count < `IDLE_DUR)
		idle_count <= idle_count + 1;
	else if (is_last_bit)
		idle_count <= 0;
end

reg pre_spi_clk_en;
always @(posedge clk or negedge reset_n)
begin
	if (~reset_n)
		pre_spi_clk_en <= 1'b0;
	else
		pre_spi_clk_en <= spi_clk_en;
end


//////////////////////////////////
// read data from SPI_MISO


// bit index
reg [11:0] bit_index;
wire is_last_bit;

assign is_last_bit = ((bit_index+1) == `BIT_NUM_PER_LINE)?1'b1:1'b0;
always @(posedge SPI_CLK or negedge reset_n)
begin
	if (~reset_n)
		bit_index <= 0;
	else if (is_last_bit)
		bit_index <= 0;
	else
		bit_index <= bit_index + 1;
end

// read data
wire [15:0] new_data;
reg [15:0]	data_queue;
assign new_data = {data_queue[14:0], SPI_MISO};


always @(posedge SPI_CLK or negedge reset_n)
begin
	if (~reset_n)
		data_queue <= 0;
	else
		data_queue <= new_data;
end

wire is_word_ready;
assign is_word_ready = ((bit_index & 12'h00F) == 12'h00F)?1'b1:1'b0;

always @(posedge SPI_CLK or negedge reset_n)
begin
	if (~reset_n)
		data <= 0;
	else if (is_word_ready) // for each 16-bit word
		data <= new_data;
end


//////////////////////////////////////
// ID and CRC Check
//reg [15:0] ID;
//reg [15:0] CRC;
reg [15:0] PRE_ID;

always @(posedge SPI_CLK or negedge reset_n)
begin
	if (~reset_n)
	begin
		ID <= 15'hFFFF;
		PRE_ID <= 15'hFFFF;
		CRC <= 15'hFFFF;
	end
	else if (bit_index == 15) // for first 16-bit word
	begin
		ID <= new_data;
		PRE_ID <= ID;
	end
	else if (bit_index == 31) // for second 16-bit word
		CRC <= new_data;
end

//////////////////////////////////////
//////////////////////////////////////
// frame control
//////////////////////////////////////
//////////////////////////////////////

/////////////////////
// frame active flag
//reg xcvr_frame;
wire check_id;
wire [11:0] line_index;

assign check_id = (bit_index == 16)?1'b1:1'b0;
assign line_index = ID[11:0];


always @(posedge SPI_CLK or negedge reset_n)
begin
	if (~reset_n)
		xcvr_frame <= 1'b0;
	else if (check_id)
	begin
		if (line_index == 12'h000 )
			xcvr_frame <= 1'b1; // first line
		else if (xcvr_frame)
		begin	
			if (line_index >= `LINE_NUM_PER_FRAME) // invalid line index
				xcvr_frame <= 1'b0;
			else if ((pre_line_index+1) != line_index) // line index out of order
				xcvr_frame <= 1'b0;
		end	
	end
	else if (xcvr_frame)
	begin
		if (is_last_bit & is_last_line)
			xcvr_frame <= 1'b0;
	end
end

//reg [15:0] frame_count;
always @(posedge SPI_CLK or negedge reset_n)
begin
	if (~reset_n)
		frame_count <= 0;
	else if (xcvr_frame)
	begin
		if (is_last_bit & is_last_line)
			frame_count <= frame_count + 1;
	end
end



reg [11:0] pre_line_index;
always @(posedge SPI_CLK or negedge reset_n)
begin
	if (~reset_n)
		pre_line_index <= 12'hFFF;
	else if (check_id)
		pre_line_index <= line_index;
end

///////////////////////
// line control
//reg [5:0] line_count;
wire is_last_line;

assign is_last_line = ((line_count+1) == `LINE_NUM_PER_FRAME)?1'b1:1'b0;
always @(posedge SPI_CLK or negedge reset_n)
begin
	if (~reset_n)
		line_count <= 0;
	else if (xcvr_frame)
	begin
		if (ID[7:0] == 0 || (PRE_ID[7:0]+1 == ID[7:0]))
			line_count <= ID[7:0];
	end
	else
		line_count <= 0;
end

*/

endmodule

